"""
Build an object to represent a Clarisse Conductor submission.

In this case the submission may contain xrefs nested to any level and we do a
pretty good job of resolving them. However, if the render errors due to xref
the localize method is the fallback.
"""

import errno
import json
import os

import shutil
import sys

from ciocore.expander import Expander
import cioclarisse.clarisse_config as ccfg
import cioclarisse.dependencies as deps
import ix

from ciocore import conductor_submit
from cioclarisse.utils import ConductorError
from cioclarisse import utils
from ciocore import data as coredata
from ciocore.sequence import Sequence
from cioclarisse import frames_ui
from cioclarisse.task import Task
from ciocore.package_environment import PackageEnvironment

from ciocore.gpath import Path
from ciocore.gpath_list import PathList


def _remove_conductor():
    """
    Remove all Conductor data from the render archive.

    This ensures the render logs are not polluted by complaints about Conductor
    nodes. This can only be done in the situation where we localize contexts,
    because in that case we get to reload the scene after submission.
    """
    objects = ix.api.OfObjectArray()
    ix.application.get_factory().get_objects("ConductorJob", objects)
    for item in list(objects):
        ix.application.get_factory().remove_item(item.get_full_name())


class Submission(object):
    """
    Submission holds all data needed for a submission.

    It has potentially many Jobs, and those Jobs each have many Tasks. A
    Submission can provide the correct args to send to Conductor, or it can be
    used to create a dry run to show the user what will happen.

    A Submission also sets a list of tokens that the user can access as <angle
    bracket> tokens in order to build strings in the UI such as commands, job
    title, and (soon to be added) metadata.
    """

    def __init__(self, node):
        """
        Collect data from the Clarisse UI.

        Collect attribute values that are common to all jobs, then call
        get_context(). After get_context has been called, the Submission level
        token variables are valid and calls to evaluate expressions will
        correctly resolve where those tokens have been used.
        """
        self.node = node
        self.project_filename = ix.application.get_current_project_filename()
        self.tmpdir = Path(utils.conductor_temp_dir())

        self.render_package_path = self._get_render_package_path()
        self.local_upload = self.node.get_attribute("local_upload").get_bool()

        self.should_delete_render_package = (
            self.node.get_attribute("clean_up_render_package").get_bool()
            and self.local_upload
        )

        self.upload_only = self.node.get_attribute("upload_only").get_bool()
        self.project = self._get_project()
        self.notifications = self._get_notifications()

        self.sequence = self._get_sequence()
        self.sources = self._get_sources()
        self.instance = self._get_instance()

        out = self._get_output_directory()
        self.common_output_path = out["common_path"]
        self.output_paths = out["output_paths"]


        tile_width = int(self.node.get_attribute("tiles").get_long())
        self.tiles = tile_width * tile_width

        sw = self._get_software()
        self.environment = sw["env"]
        self.package_ids = sw["ids"]


        self.dependencies = deps.collect(self.node)
        self.dependencies.add(self.render_package_path)
 

        self.tokens = self.get_context()
        expander = Expander(**self.tokens)
        self.title = expander.evaluate(self.node.get_attribute("title").get_string())

 
        self.tasks = []

        task_att = self.node.get_attribute("task_template")
        for chunk in self.sequence["main"].chunks():
            for tile_number in range(1, self.tiles + 1):
                tile_spec = (self.tiles, tile_number)
                self.tasks.append(
                    Task(chunk, task_att, self.sources, tile_spec, self.tokens)
                )



    def _get_sources(self):
        """
        Get the images/layers, along with associated Sequence objects.

        If we are not rendering a custom range, then the sequence for
        each image may be different.

        Returns:
            list of dict: elements contain an image along with the Sequence that
            represents the image range.
        """

        images = ix.api.OfObjectArray()
        self.node.get_attribute("images_and_layers").get_values(images)

        use_custom = self.node.get_attribute("use_custom_frames").get_bool()

        # cast to list because OfObjectArray is true even when empty.
        if not list(images):
            msg = "No render outputs added. Please add images and/or layers in the Submitter"
            raise ConductorError(msg)
        seq = self.sequence["main"]
        result = []
        for image in images:
            if not use_custom:
                seq = Sequence.create(*frames_ui.image_range(image))
            result.append({"image": image, "sequence": seq})
        return result


    def _get_sequence(self):

        """
        Get sequence objects from the frames section of the UI.

        Returns:
            dict: main sequence and scout sequence.
        """
        return {
            "main": frames_ui.main_frame_sequence(self.node),
            "scout": frames_ui.resolved_scout_sequence(self.node),
        }



    def _get_instance(self):
        """
        Get everything related to the instance.

        Get the machine type, preemptible flag, and number of retries if
        preemptible. We use the key from the instance_type menu and look
        up the machine spec in the shared data where the full list of
        instance_types is stored. When exhaustion API is in effect, the
        list of available types may be dynamic, so wetell the user to
        refresh.

        Returns:
            dict: Fields to specify the render node behaviour.
        """
        instance_types = coredata.data().get("instance_types")

        label = self.node.get_attribute("instance_type").get_applied_preset_label()

        result = {
            "preemptible": self.node.get_attribute("preemptible").get_bool(),
            "retries": self.node.get_attribute("retries").get_long(),
        }

        try:
            found = next(it for it in instance_types if str(it["description"]) == label)
        except StopIteration:
            msg = 'Invalid instance type "{}". Try a refresh (Press the Connect button).'.format(label)
            raise ConductorError(msg)

        result.update(found)
        return result



    def _get_project(self):
        """Get the project from the attr.

        Get its ID in case the current project is no longer in the list
        of projects at conductor, throw an error.
        """


        projects = coredata.data().get("projects")
        project_att = self.node.get_attribute("conductor_project_name")
        label = project_att.get_applied_preset_label()

        found = next((p for p in projects if p== label), None)
        if not found:
            msg = 'Cannot find project "{}" at Conductor. Please ensure the PROJECT dropdown contains a valid project.'.format(label)
            raise ConductorError(msg)

        return found

    def _get_notifications(self):
        """Get notification prefs."""
        if not self.node.get_attribute("notify").get_bool():
            return None

        emails = self.node.get_attribute("email_addresses").get_string()
        return [email.strip() for email in emails.split(",") if email.strip()]

    def get_context(self):
        """Env tokens are variables to help the user build expressions.

        The user interface has fields for strings such as job title,
        task command. The user can use these tokens with <angle brackets> to build those strings. Tokens at the Submission
        level are also available in Job level fields, and likewise
        tokens at the Job level are available in Task level fields.
        """
        tokens = {}

        pdir_val = ix.application.get_factory().get_vars().get("PDIR").get_string()

        tokens["pdir"] = '"{}"'.format(Path(pdir_val).posix_path(with_drive=False))

        tokens["temp_dir"] = "{}".format(self.tmpdir.posix_path(with_drive=False))
        tokens["tempdir"] =tokens["temp_dir"]

        tokens["submitter"] = self.node.get_name()

        tokens["render_package"] = '"{}"'.format(
            self.render_package_path.posix_path(with_drive=False)
        )
        tokens["renderpackage"] = tokens["render_package"] 

        tokens["project"] = self.project 

        main_seq = self.sequence["main"]
        scout_seq = self.sequence["scout"]

        tokens["scout"] = str(scout_seq)
        tokens["chunksize"] = str(main_seq.chunk_size)
        tokens["chunkcount"] = str(main_seq.chunk_count())
        tokens["scoutcount"] = str(len(scout_seq or []))
        tokens["sequencelength"] = str(len(main_seq))
        tokens["sequence"] = str(main_seq)
        tokens["sequencemin"] = str(main_seq.start)
        tokens["sequencemax"] = str(main_seq.end)
        tokens["instance_type"] = self.instance["name"]
        tokens["instance"] = self.instance["description"]
        tokens["tiles"] = str(self.tiles)
        pidx = int(self.instance["preemptible"])
        tokens["preemptible"] = "preemptible" if pidx else "non-preemptible"
        tokens["retries"] = str(self.instance["retries"])
        tokens["job"] = self.node_name

        # Space delimited list of output paths are needed for a mkdir cmd. As
        # mentioned above, no longer needed now that directories are created in
        # the prerender script. Just don't want to remove right now as no time
        # to test.
        tokens["directories"] = " ".join(
            '"{}"'.format(p.posix_path(with_drive=False)) for p in self.output_paths
        )

        return tokens

    def _get_render_package_path(self):
        """
        Calc the path to the render package.

        The name is not always known until
        preview/submission time because it is based on the filename. 

        It will however always show up in the preview window.

        We replace spaces in the filename because of a bug in Clarisse
        https://www.isotropix.com/user/bugtracker/376

        Returns:
            string: path
        """

        msg= 'Cannot create a submission from this file. Has it ever been saved?'
        current_filename = ix.application.get_current_project_filename()
        if not current_filename:
            raise ConductorError(msg)

        node_name =self.node.get_attribute("title").get_string()

        node_name = "".join(x for x in node_name if x.isalpha() or x.isdigit()  )


        path = os.path.splitext(current_filename)[0]

        path = os.path.join(
            os.path.dirname(path), os.path.basename(path).replace(" ", "_")
        )
 
        try:
            result =  Path("{}.{}.cio.project".format(path,node_name))
        except ValueError as ex:
            msg = "{} - while resolving {}".format(str(ex), result)
            raise ConductorError(msg)
        
        return result




    def _get_output_directory(self):
        """
        Get the common path for all image output paths.

        NOTE: We don't really need the subpaths any longer because directory
        creation is handled in the prerender script. Don't want to mess with
        things right now though.

        Returns:
            dict: common path and list of paths below it
        """
        out_paths = PathList()

        images = ix.api.OfObjectArray()
        self.node.get_attribute("images_and_layers").get_values(images)

        for image in images:
            directory = os.path.dirname(image.get_attribute("save_as").get_string())
            try:
                out_paths.add(directory)
            except ValueError as ex:
                msg = "{} - while resolving {}".format(str(ex), directory)
                raise ConductorError(msg)
        return {"common_path": out_paths.common_path(), "output_paths": out_paths}


 


    def _get_software(self):
        
        result = {
            "ids": [],
            "env": PackageEnvironment()
        }

        clarisse_version_att = self.node.get_attribute("clarisse_version")
        name = clarisse_version_att.get_applied_preset_label()

        tree_data = coredata.data().get("software")
        host_package = tree_data.find_by_path(name)

        result["ids"].append(host_package["package_id"])
        result["env"].extend(host_package)
        result["env"].extend(self._get_internal_extra_vars())
        result["env"].extend(self._get_extra_env_vars())

        return result


    def _get_extra_env_vars(self):
        """
        Collect any environment specified by the user.

        Returns:
            list of dict: name, value and merge policy of user specified env vars.
        """

        result = []
        json_entries = ix.api.CoreStringArray()
        self.node.get_attribute("extra_environment").get_values(json_entries)

        for entry in [json.loads(j) for j in json_entries]:
            result.append(
                {
                    "name": entry["key"],
                    "value": os.path.expandvars(entry["value"]),
                    "merge_policy": ["append", "exclusive"][int(entry["excl"])],
                }
            )

        return result

    def _get_internal_extra_vars(self):
        return [
            {
                "name": "PYTHONHOME",
                "value": "/opt/silhouettefx/silhouette/7/silhouette-7.5.2",
                "merge_policy": "exclusive",
            },
            {"name": "CONDUCTOR_PATHHELPER", "value": 0, "merge_policy": "exclusive"},
            {
                "name": "LD_LIBRARY_PATH",
                "value": "/usr/lib/python2.7/config-x86_64-linux-gnu",
                "merge_policy": "append",
            },
        ]

    def get_args(self):
        """
        Prepare the args for submission to conductor.

        Returns:
            list: list of dicts containing submission args per job.
        """
        result = {}

        result["local_upload"] = self.local_upload
        result["upload_only"] = self.upload_only
        
        result["project"] = self.project
        result["notify"] = self.notifications

        result["upload_paths"] = sorted([d.posix_path() for d in self.dependencies])
        result["autoretry_policy"] = (
            {"preempted": {"max_retries": self.instance["retries"]}}
            if self.instance["preemptible"]
            else {}
        )
        result["software_package_ids"] = self.package_ids 

        result["preemptible"] = self.instance["preemptible"]
        result["environment"] = dict(self.environment)
        result["scout_frames"] = ", ".join(
            [str(s) for s in self.sequence["scout"] or []]
        )
        result["output_path"] = self.common_output_path.posix_path()
        result["chunk_size"] = self.sequence["main"].chunk_size
        result["instance_type"] = self.instance["name"]
        if not result["upload_only"]:
            result["tasks_data"] = [task.data() for task in self.tasks]

        result["job_title"] = self.title
  
        return result

    def submit(self):
        """
        Submit all jobs.

        Returns:
            list: list of response dictionaries, containing response codes
            and descriptions.
        """

        submission_args = self.get_args()
        self._before_submit()

        self.remove_missing_upload_paths(submission_args)

        try:
            remote_job = conductor_submit.Submit(submission_args)
            response, response_code = remote_job.main()
            ix.log_info({"code": response_code, "response": response})
        except BaseException:
            raise
        finally:
            self._after_submit()
        return response

    def _before_submit(self):
        """
        Prepare the project files that will be shipped.

        We first write out the current project file. 
        
        Then (on Windows) we find additional referenced project 
        files and adjust paths in all of them so they may be 
        rendered on linux render nodes.
        """
        self.write_render_package()


    def write_render_package(self):
        """
        Write a package suitable for rendering.

        A render package is a project file with a special name.
        """

        self._before_write_package()

        context = ix.get_item("project:/")
        ix.log_info("Writing render package: context is '{}'".format(context))
            
        package_file = self.render_package_path.posix_path()

        with utils.disabled_app():
            success = ix.export_context_as_project_with_dependencies(context, package_file)
        
        if success:
            ix.log_info("Wrote render package '{}'".format(package_file))
        else:
            msg = "Failed to write render package file '{}'".format(package_file)
            raise ConductorError(msg)

        self._after_write_package()

        if utils.is_windows():
            ix.log_info("Windows path adjustments")
            utils.linuxify(self.render_package_path.posix_path())
            ix.log_info("Linuxified render project file")
        else:
            ix.log_info("Not using Windows")

        return package_file

 

    def remove_missing_upload_paths(self, submission_args):
        """
        Alert the user of missing files. If the user doesn't want to continue
        with missing files, the result will be False. Otherwise it will be True
        and the potentially adjusted args are returned.

        Args:
            submission_args (obj): 

        Returns:
           tuple (bool, adjusted args):
        """
        missing_files = []

  
        existing_files = []
        for path in submission_args["upload_paths"]:
            if os.path.exists(path):
                existing_files.append(path)
            else:
                missing_files.append(path)

        submission_args["upload_paths"] = existing_files

        missing_files = sorted(list( missing_files))
        if missing_files:
            ix.log_warning("Skipping missing files:")
            for f in missing_files:
                ix.log_warning(f)


    def _before_write_package(self):
        """
        Prepare to write render package.
        """
        self._prepare_temp_directory()
        self._copy_system_dependencies_to_temp()

    def _prepare_temp_directory(self):
        """
        Make sure the temp directory has a conductor subdirectory.
        """
        tmpdir = self.tmpdir.posix_path()
        try:
            os.makedirs(tmpdir)
        except OSError as ex:
            if not (ex.errno == errno.EEXIST and os.path.isdir(tmpdir)):
                raise
        ix.log_info("Prepared tmpdir '{}'".format(tmpdir))

    def _copy_system_dependencies_to_temp(self):
        """
        Copy over all system dependencies to a tmp folder.

        Wrapper scripts, config files etc. The clarisse.cfg file is special. See
        ../clarisse_config.py
        """
        for entry in deps.system_dependencies():
            if os.path.isfile(entry["src"]):
                if entry["src"].endswith(".cfg"):
                    safe_config = ccfg.legalize(entry["src"])
                    with open(entry["dest"], "w") as dest:
                        dest.write(safe_config)
                    ix.log_info(
                        "Copy with mods {} to {}".format(entry["src"], entry["dest"])
                    )
                else:
                    ix.log_info("Copy {} to {}".format(entry["src"], entry["dest"]))
                    shutil.copy(entry["src"], entry["dest"])


    def _after_submit(self):
        """Clean up, and potentially other post submission actions."""
        self._delete_render_package()

    def _after_write_package(self):
        """
        Runs operations after saving the render package.

        If we did something destructive, like localize contexts, then
        a backup will have been saved and we now reload it. This strategy
        is used because Clarisse's undo is broken when it comes to
        undoing context localization.
        """
        pass
 
    def _delete_render_package(self):
        """
        Delete the render package from disk if the user wants to.
        """
        if self.should_delete_render_package:
            render_package_file = self.render_package_path.posix_path()
            if os.path.exists(render_package_file):
                os.remove(render_package_file)
 

    @property
    def node_name(self):
        """node_name."""
        return self.node.get_name()

    @property
    def filename(self):
        """filename."""
        return ix.application.get_current_project_filename()

    def has_notifications(self):
        """has_notifications."""
        return bool(self.notifications)

    @property
    def email_addresses(self):
        """email_addresses."""
        if not self.has_notifications():
            return []
        return self.notifications["email"]["addresses"]



# import abc
# import importlib
# import inspect
import os
import ix

from ciocore.gpath import Path
from ciocore.gpath_list import GLOBBABLE_REGEX, PathList
from ciocore.validator import Validator, ValidationError
from cioclarisse import const as k
from ciocore import data as coredata
from cioclarisse import frames_ui 
from cioclarisse import dependencies as deps
from cioclarisse import validation_ui
from cioclarisse import utils

from ciocore.sequence import Sequence
 
class ValidateImagesAndLayers(Validator):
    def run(self, _):
 
        images = ix.api.OfObjectArray()
        self._submitter.get_attribute("images_and_layers").get_values(images)
        out_paths = PathList()

        num_images = images.get_count()

        if not num_images:
            msg = "No render outputs added. Please add images and/or layers in the Submitter"
            self.add_error(msg)
            return

        for image in images:
            if not image.get_attribute("render_to_disk").get_bool():
                msg = "Image or Layer does not have render_to_disk attribute set: {}".format(image.get_full_name())
                self.add_error(msg)
                continue
            

            save_path = image.get_attribute("save_as").get_string()
            if not save_path:
                msg = "Image save_as path is not set: {}".format(image.get_full_name())
                self.add_error(msg)
                continue
    
            if save_path.endswith("/"):
                msg =  "Image save_as path must be a filename, not a directory: {}".format( image.get_full_name() )
                self.add_error(msg)
                continue
    

            directory = os.path.dirname(save_path)
            out_paths.add(directory)

        if not out_paths:
            msg = "There are no images or layers with Save to Disk turned on."
            self.add_error(msg)
            return
    
        common_path = out_paths.common_path()

        paths = "\n".join(p.posix_path() for p in out_paths)

        if common_path.depth == 0:
            msg = "Your output files should be rendered to a common folder. {}\n{}".format(common_path.posix_path(), paths )
            self.add_error(msg)
            return

        if self._submitter.get_attribute("use_custom_frames").get_bool():
            frame_seq = frames_ui.custom_frame_sequence(self._submitter)
            if not frame_seq:
                self.add_error("Error validating custom frames sequence.")
            for image in images:
                image_rng = frames_ui.image_range(image)
                image_seq = Sequence.create(*image_rng)
                isect = image_seq.intersection(frame_seq)
                if list(isect) != list(frame_seq):
                    self.add_warning("Image has less frames than custom sequence. {}".format( image.get_full_name()))


class ValidateTaskCount(Validator):
    def run(self, _): 
        seq = frames_ui.main_frame_sequence(self._submitter)
        if not seq:
            self.add_error("No valid frame seqence. Either ensure connected images have a valid sequence, or overide the fame range in the subitter.")
            return
        count = seq.chunk_count()
        if  count > 1000:
            self.add_notice(
                "This submission contains over 1000 tasks ({}). Are you sure this is correct?".format(count))

class ValidatePackages(Validator):
    def run(self, _):
 
        if not (
            self._submitter.get_attribute("clarisse_version")
            .get_applied_preset_label()
            .startswith("clarisse")
        ):
            self.add_error("No Clarisse package selected.")

class ValidateProject(Validator):
    def run(self, _): 
        projects = coredata.data().get("projects")

        project_att = self._submitter.get_attribute("conductor_project_name")
        label = project_att.get_applied_preset_label()
        if not label or label == k.NOT_CONNECTED :
            msg = 'Project is not set for "{}".'.format(self._submitter.get_name())
            self.add_error(msg)
        try:
            next(p for p in projects if str(p) == label)
        except StopIteration:
            msg = 'Cannot find project "{}" at Conductor. Please ensure the PROJECT dropdown contains a valid project.'.format(label)
            self.add_error(msg)
 
class ValidateUploadDaemon(Validator):


    def run(self, _): 
        if  self._submitter.get_attribute("local_upload").get_bool():
            msg ="Local Upload is turned on, which means your Clarisse session will be\n"
            msg+="unusable until all assets have uploaded. If you have a large number of\n"
            msg+="assets that don't already exist on Conductor, you are advised to use a\n"
            msg+="background process to handle the upload. Its easy to set up. Just cancel\n"
            msg+="this window, turn Local Upload OFF in the submitter and resubmit. You'll\n"
            msg+="then see instructions to start the upload daemon."

            self.add_notice(msg)
            return
        # daemon
        msg  = "Local Upload is turned off. Therefore this submission expects an uploader\n"
        msg += "daemon to be running. If you haven't started an uploader daemon, follow\n"
        msg += "the instructions below.\n"
        if utils.is_windows():
            msg += "Open a new command prompt or Powershell and type:\npython.exe Conductor\\bin\\conductor uploader\nthen press enter."
        else:
            msg += "Open a new terminal window and type:\nConductor/bin/conductor uploader\nthen press enter."
        msg += "For more information, search 'upload daemon' at docs.conductortech.com"
        self.add_notice(msg)

class ValidateMissingAssets(Validator):

    def run(self, _):
 
        basenames =  deps.CONDUCTOR_SCRIPTS + [deps.CLARISSE_CFG_FILENAME]

        path_list = PathList()

        for gpath in deps.collect(self._submitter, do_glob=False):

            # we can assunme that all globbable paths, (those containing glob
            # characters) cannot possibly represent missing files because they are
            # literally resolved  by checking what matches on disk. Therfore, we ignore them.
            if not GLOBBABLE_REGEX.search(gpath.posix_path()):
                path_list.add(gpath)

        # path_list has taken care of deduplication by now - yay
        missing = []
        for gpath in path_list:
            pp = gpath.posix_path()
            if any(pp.endswith(x) for x in basenames):
                continue
            if not os.path.exists(pp):
                missing.append(pp)

        if missing:
            for asset in missing:
                self.add_warning("Missing: {}".format(asset))

def run(node):
 
    errors, warnings, notices = _run_validators(node)

    if errors:
        for error in errors:
            ix.log_error(error)
        msg = "There are some critical issues. You can't submit until you fix them."
        raise ValidationError(msg)

    if notices or warnings:
        dialog_result = validation_ui.proceed(warnings,notices)
        
        if not dialog_result:
            msg = "Submission cancelled by user."
            raise ValidationError(msg)
    else:
        ix.log_info("There are no Notices or Warnings")

 
def _run_validators(node):
    validators =  [plugin(node) for plugin in Validator.plugins()]
    for validator in validators:
        validator.run("main")

    errors = list(set.union(*[validator.errors for validator in validators]))
    warnings = list(
        set.union(*[validator.warnings for validator in validators]))
    notices = list(set.union(*[validator.notices for validator in validators]))
    return errors, warnings, notices



"""
Responds to events in the clarisse version section of the ConductorJob attribute editor.
"""

from ciocore import data as coredata
from cioclarisse import const as k
def update(obj):
    """
    Rebuilds the clarisse_version menu.

    Args:
        obj (ConductorJob): Item on which to rebuild menu.
    """

    software_att = obj.get_attribute("clarisse_version")
    if not coredata.valid():
        software_att.add_preset(k.NOT_CONNECTED,  "0")
        return

    software_data = coredata.data().get("software")
    for i, name in enumerate(software_data.supported_host_names()):
        software_att.add_preset(name.encode("utf-8"), str(i))



"""
Module is responsible for providing documentation for all attributes of the scripted class.
"""


def set_docs(s_class):

    s_class.set_attr_doc(
        "connect",
        """Connects to Conductor and updates available projects and machine types. Sign in may be required. Also sets empty string attributes to their defaults.""",
    )

    s_class.set_attr_doc(
        "preview",
        """Build a submission as json and inspect it before submitting to Conductor.""",
    )

    s_class.set_attr_doc("submit", """Submits this job to Conductor.""")

    s_class.set_attr_doc(
        "title",
        """Sets the title that appears in the Conductor dashboard. You may use Clarisse variables and some Conductor tokens to construct the title.""",
    )

    s_class.set_attr_doc(
        "images_and_layers",
        """Sets the items to be rendered. Images or Layers must have their Render to Disk attribute set. The Save As field must contain a filename.""",
    )

    s_class.set_attr_doc("conductor_project_name", """Sets the Conductor project.""")

    s_class.set_attr_doc(
        "use_custom_frames",
        """Overrides the frame ranges from connected image nodes.""",
    )

    s_class.set_attr_doc(
        "custom_frames",
        """Specifies a set of frames to render. Provide a comma-separated list of progressions. Example 1,7,10-20,30-60x3,1001.""",
    )

    s_class.set_attr_doc("chunk_size", """Sets the number of frames per task.""")

    s_class.set_attr_doc("use_scout_frames", """Activates a set of frames to scout.""")

    s_class.set_attr_doc(
        "scout_frames",
        """Scout-frames to render. Tasks containing any of the specified scout frames, are run in their entirety. Other frames are set to a holding state.""",
    )

    s_class.set_attr_doc(
        "tiles", """Render tiles. Split each single frame across many machines."""
    )

    s_class.set_attr_doc(
        "preemptible",
        """Activates use of low-cost instances. Premptble instances may be stopped at any time by the cloud provider. See the Conductor documentation site for a discussion of the situations where preemptible instances are suitable.""",
    )

    s_class.set_attr_doc(
        "instance_type", """Specifies the required hardware configuration."""
    )

    s_class.set_attr_doc(
        "retries",
        """Sets the number of times to retry a failed task before marking it failed.""",
    )

    s_class.set_attr_doc(
        "dependency_scan_policy",
        """Specifies how to search the project for references to files it depends on. The Smart Sequence option tries to identify only those files that are needed for the specified frames. The Glob option finds any files that match a hash pattern. If you have saved all required files in the cached upload list, you can set the policy to No Scan.""",
    )

    s_class.set_attr_doc(
        "local_upload",
        """Uploads files from your Clarisse session, as opposed to using an upload daemon.""",
    )

    s_class.set_attr_doc(
        "upload_only", """Uploads files but does not start any tasks."""
    )

    s_class.set_attr_doc(
        "manage_extra_uploads",
        """Opens a panel to browse or scan for files to upload. If any files are not found by the dependency scan at submission time, they may be added here.""",
    )

    s_class.set_attr_doc(
        "extra_uploads",
        """Files to uploaded in addition to any files found by dependency scanning.""",
    )

    s_class.set_attr_doc(
        "clarisse_version",
        """Choose the version of Clarisse to run on the render nodes.""",
    )

    s_class.set_attr_doc(
        "manage_extra_environment",
        """Opens a panel for making modifications to the remote environment.""",
    )

    s_class.set_attr_doc(
        "extra_environment", """Extra environment encoded as a JSON string."""
    )

    s_class.set_attr_doc(
        "task_template",
        """Specifies a template for the command that will be run on remote instances. See the Conductor documentation site for a detailed discussion.""",
    )

    s_class.set_attr_doc(
        "localize_contexts",
        """Localize contexts before dispatching to Conductor. This will cause your project to be saved and then reloaded after submission, which may take a long time for large projects.""",
    )

    s_class.set_attr_doc(
        "clean_up_render_package",
        """Removes the render project file after submission. If local_upload is switched off, then this attribute will be hidden and the render project will not be removed.""",
    )

    s_class.set_attr_doc(
        "notify",
        """Indicates that notifications will be sent by email on job completion.""",
    )

    s_class.set_attr_doc(
        "email_addresses",
        """A comma delimited list of emails addresses to notify on job completion.""",
    )

    s_class.set_attr_doc(
        "show_tracebacks", """Shows a stacktrace when something goes wrong."""
    )

    s_class.set_attr_doc(
        "conductor_log_level",
        """Sets the severity above which Conductor's logging emits messages.""",
    )



"""
Module concerned with refreshing the datablock and the attribute editor.
"""

import ix
from cioclarisse import (
    software_ui,
    debug_ui,
    frames_ui,
    instances_ui,
    projects_ui,
)

from ciocore import data as coredata

def force_ae_refresh(node):
    """
    Trigger an attribute editor refresh.

    Args:
        node (ConductorJob): item whose attribute editor to refresh.
    """
    attr = node.get_attribute("instance_type")
    count = attr.get_preset_count()
    applied = attr.get_applied_preset_index()
    if count:
        last = count - 1
        preset = (attr.get_preset_label(last), attr.get_preset_value(last))
        attr.remove_preset(last)
        attr.add_preset(preset[0], preset[1])
        attr.set_long(applied)


def refresh(_, **kw):
    """
    Respond to connect button click, and others that may benefit from a refresh.

    Update UI for projects and instances from the data block.

    Args:
        (kw["force"]): If true, invalidate the datablock and fetch fresh from
        Conductor. We update all ConductorJob nodes. Also update the log level
        in the UI.
    """
    try:
        coredata.data(force=True)
    except BaseException as ex:

        ix.log_error("""
Can't get data from Conductor. Try the connect button again. 
If that doesn't work, your credentials file may be corrupt.
Please delete it from here (~/.config/conductor/credentials) and try again.
            """)
        raise
        

    nodes = ix.api.OfObjectArray()
    ix.application.get_factory().get_all_objects("ConductorJob", nodes)

    for node in nodes:
        projects_ui.update(node)
        instances_ui.update(node)
        software_ui.update(node)
        frames_ui.update_frame_stats_message(node)

    debug_ui.refresh_log_level(nodes)



"""
Responds to events in the projects UI.
"""
from ciocore import data as coredata
from cioclarisse import const as k

def handle_project(obj, attr):
    """
    When project changes, stash its name on a string att.

    The current applied option is saved by its index. There may be a
    case where the index that was saved in the file is different from
    the actual project that the index referred to when the scene was
    saved, if new projects are added with a lower alphabetical position
    for example. In this case we can look at the stashed project name
    and figure out what to do.

    NOTE: the reassignment part is not yet
    implemented, but the stashed name part is.

    Args:
        obj (ConductorJob):
        attr (OfAttr): Attribute that changed.
    """
    label = attr.get_applied_preset_label()
    if label:
        obj.get_attribute("last_project").set_string(label)


def update(obj):
    """
    Fetch projects afresh and repopulate menu.  

    Args:
        obj (ConductorJob):
     """

    project_att = obj.get_attribute("conductor_project_name")
    project_att.remove_all_presets()

    if not coredata.valid():
        project_att.add_preset(k.NOT_CONNECTED,  "0")
        return

    projects = coredata.data().get("projects")
    for i, project in enumerate(projects):
        project_att.add_preset(project.encode("utf-8"), str(i))



"""
Responds to events in the  upload section of the ConductorJob attribute editor.
"""

import ix


def handle_local_upload(obj, attr):
    """
    Responds to local_upload changes.

    We must not clean up the render file if local_upload is false,
    because it cleans up before the upload daemon has had a chance
    to upload it.

    Args:
        obj (ConductorJob):
        attr (OfAttr): Attribute that changed.
    """
    obj.get_attribute("clean_up_render_package").set_hidden(not attr.get_bool())



"""
Contexts for performing long or delicate functions.
"""
import os
import re
from contextlib import contextmanager
import tempfile
import shutil
import ix
from ciocore.gpath import Path


FILE_ATTR_HINTS = [
    ix.api.OfAttr.VISUAL_HINT_FILENAME_SAVE,
    ix.api.OfAttr.VISUAL_HINT_FILENAME_OPEN,
    ix.api.OfAttr.VISUAL_HINT_FOLDER,
]


class ConductorError(Exception):
    pass


@contextmanager
def waiting_cursor():
    """
    Perform some function with the wait cursor showing.
    """
    clarisse_win = ix.application.get_event_window()
    old_cursor = clarisse_win.get_mouse_cursor()
    clarisse_win.set_mouse_cursor(ix.api.Gui.MOUSE_CURSOR_WAIT)
    yield
    clarisse_win.set_mouse_cursor(old_cursor)


@contextmanager
def disabled_app():
    """
    Disble the app to perform some function.
    """
    app = ix.application
    app.disable()
    yield
    app.enable()


def is_windows():
    return os.name == "nt"


def conductor_temp_dir():
    return os.path.join(
        ix.application.get_factory().get_vars().get("CTEMP").get_string(),
        "conductor",
    )


def linuxify(filename):
    """
    Adjust reference pasths for windows.
    """
    path_regex = _get_path_line_regex()

    temp_path =  tempfile.mktemp()

    shutil.copy2(filename, temp_path)
    os.remove(filename)
    with open(filename, "w+") as outfile:
        with open(tempfile, "r+") as infile:
            for line in infile:
                match = re.match(path_regex, line)
                if match:
                    path = Path(match.group(1), no_expand=True).posix_path(with_drive=False)
                    outfile.write(line.replace(match.group(1), path))
                else:
                    outfile.write(line)



def _get_path_line_regex():
    """
    Generate a regex to help identify filepath attributes.

    As we scan project files to replace windows paths, we use this regex which
    will be something like: r'\s+(?:filename|filename_sys|save_as)\s+"(.*)"\s+' 
    only longer.
    """
    classes = ix.application.get_factory().get_classes()
    file_attrs = []
    for klass in classes.get_classes():
        attr_count = klass.get_attribute_count()
        for i in xrange(attr_count):
            attr = klass.get_attribute(i)
            hint = attr.get_visual_hint()
            if hint in FILE_ATTR_HINTS:
                file_attrs.append(attr.get_name())

    return r"\s+(?:" + "|".join(sorted(set(file_attrs))) + r')\s+"(.*)"\s+'

 


"""
Provides a window in which to add extra uploads.

This is required because Clarisse's attribute editor doesn't allow
custom UI to be embedded.
"""
import cioclarisse.dependencies as deps
import ix
from cioclarisse import refresh
from ciocore.gpath_list import PathList
from cioclarisse.utils import ConductorError
BTN_HEIGHT = 22
BTN_WIDTH = 100
WINDOW_LEFT = 600
WINDOW_TOP = 200
HEIGHT = 300
WIDTH = 900


class FileListWidget(ix.api.GuiTree):
    """
    Build a list of files using a tree widget.

    Use the tree widget because its the only way to get multiselect in a
    clarisse list-like widget. The actual contents will be flat, one level deep.
    """

    def __init__(self, parent, y_val):
        """
        Set up the tree config.

        We inherit so that we can keep a list of children, due to
        GuiWidget,get_child_items being broken.

        Args:
            parent (GuiWidget): Parent UI
            y_val (int): Height location to put the top of the widget
        """

        ix.api.GuiTree.__init__(self, parent, 0, y_val, WIDTH, HEIGHT)
        self.set_border_style(ix.api.GuiTree.BORDER_STYLE_FLAT)
        self.set_shaded_entries(True)
        self.enable_multi_selection(True)
        self.enable_item_delete(True)
        self.item_list = []

    def add_entries(self, entries):
        """
        Add a line item for each entry.

        Use PathList to deduplicate on the fly. As the addition of
        entries may completely change the list (grow or shrink) we
        delete and rebuild the list of entries each time a batch is
        added.

        Args:
            entries (list): list of file paths.
        """
        if not entries:
            return

        root_item = self.get_root()

        deduped = PathList()
        for item in self.item_list:
            try:
                deduped.add(item.get_name())
            except ValueError as ex:
                msg = "{} - while resolving {}".format(str(ex), item.get_name())
                raise ConductorError(msg)

        for entry in entries:
            try:
                deduped.add(entry)
            except ValueError as ex:
                
                msg = "{} - while resolving {}".format(str(ex), entry)
                raise ConductorError(msg)
 
        # clear existing list
        root_item.remove_children()
        del self.item_list[:]

        for entry in deduped:
            item = ix.api.GuiTreeItemBasic(root_item, entry.posix_path())
            self.item_list.append(item)
        self.refresh()

    def destroy_selected(self):
        """
        Remove the selected items.

        Also remove from the item_list.
        """
        indices = self._get_selected_indices()
        if not indices:
            ix.log_warning("Nothing selected")
            return
        self.destroy_selected_items()
        self.refresh()
        self._remove_from_sync_list(indices)

    def destroy_all(self):
        """
        Clear all items.
        """
        self.get_root().remove_children()
        del self.item_list[:]
        self.refresh()

    def _get_selected_indices(self):
        """
        Get indices of selected paths.

        Get in reverse order so they don't change during removal.
        """
        indices = sorted(
            [i for i, item in enumerate(self.item_list) if item.is_selected()]
        )
        indices.reverse()
        return indices

    def _remove_from_sync_list(self, indices):
        """
        Remove items with given indices.

        Args:
            indices (list): Indices of items to remove.
        """
        for index in indices:
            self.item_list.pop(index)


class ExtraUploadsWindow(ix.api.GuiWindow):
    """
    Window to contain the tree.

    It has buttons to browse and manage the list. It also has regular apply, go,
    close buttons at the bottom.
    """

    def __init__(self, node):

        self.node = node

        window_height = HEIGHT + (BTN_HEIGHT * 2)
        ix.api.GuiWindow.__init__(
            self,
            ix.application.get_event_window(),
            WINDOW_LEFT,
            WINDOW_TOP,
            WIDTH,
            window_height,
            "Extra uploads",
        )
        self.set_resizable(False)

        current_y = 0
        self.browse_but = ix.api.GuiPushButton(
            self, (WIDTH - BTN_WIDTH), current_y, BTN_WIDTH, BTN_HEIGHT, "Browse"
        )
        self.connect(self.browse_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_browse_but)

        # self.smart_scan_but = ix.api.GuiPushButton(
        #     self,
        #     (WIDTH - (BTN_WIDTH * 2)),
        #     current_y,
        #     BTN_WIDTH,
        #     BTN_HEIGHT,
        #     "Smart scan",
        # )
        # self.connect(
        #     self.smart_scan_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_smart_scan_but
        # )

        # self.glob_scan_but = ix.api.GuiPushButton(
        #     self,
        #     (WIDTH - (BTN_WIDTH * 3)),
        #     current_y,
        #     BTN_WIDTH,
        #     BTN_HEIGHT,
        #     "Glob scan",
        # )
        # self.connect(
        #     self.glob_scan_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_glob_scan_but
        # )

        self.remove_sel_but = ix.api.GuiPushButton(
            self, 0, current_y, BTN_WIDTH, BTN_HEIGHT, "Remove selected"
        )
        self.connect(
            self.remove_sel_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_remove_sel_but
        )

        self.remove_all_but = ix.api.GuiPushButton(
            self, BTN_WIDTH, current_y, BTN_WIDTH, BTN_HEIGHT, "Remove all"
        )
        self.connect(
            self.remove_all_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_remove_all_but
        )

        current_y += BTN_HEIGHT
        self.file_list_wdg = FileListWidget(self, current_y)
        current_y += HEIGHT

        go_btn_width = int(WIDTH / 3)

        self.close_but = ix.api.GuiPushButton(
            self, 0, current_y, go_btn_width, BTN_HEIGHT, "Close"
        )
        self.connect(self.close_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_close_but)

        self.apply_but = ix.api.GuiPushButton(
            self, go_btn_width, current_y, go_btn_width, BTN_HEIGHT, "Apply"
        )
        self.connect(self.apply_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_apply_but)

        self.go_but = ix.api.GuiPushButton(
            self, (WIDTH - go_btn_width), current_y, go_btn_width, BTN_HEIGHT, "Go"
        )
        self.connect(self.go_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_go_but)

    def on_browse_but(self, sender, eventid):
        """
        User may browse for files.
        """
        app = ix.application
        title = "Select extra uploads..."
        mask = "Any files\t*"
        filenames = ix.api.GuiWidget.open_files(app, "", title, mask)
        self.file_list_wdg.add_entries(sorted(filenames))

    # def on_smart_scan_but(self, sender, eventid):
    #     """
    #     Add the results of a smart scan to the list.
    #     """
    #     include_references = not self.node.get_attribute("localize_contexts").get_bool()
    #     filenames = deps.get_scan(self.node, deps.SMART, include_references)
    #     filenames.glob()
    #     self.file_list_wdg.add_entries(sorted(filenames))

    # def on_glob_scan_but(self, sender, eventid):
    #     """
    #     Add the results of a glob scan to the list.
    #     """
    #     include_references = not self.node.get_attribute("localize_contexts").get_bool()
    #     filenames = deps.get_scan(self.node, deps.GLOB, include_references)
    #     filenames.glob()
    #     self.file_list_wdg.add_entries(sorted(filenames))

    def on_remove_sel_but(self, sender, eventid):
        """
        Responds to remove selected button.
        """
        self.file_list_wdg.destroy_selected()

    def on_remove_all_but(self, sender, eventid):
        """
        Responds to remove all button.
        """
        self.file_list_wdg.destroy_all()

    def on_close_but(self, sender, eventid):
        """
        Hide the widget.

        Destruction will be be triggered as the event loop ends.
        """
        self.hide()

    def on_apply_but(self, sender, eventid):
        self._apply()

    def on_go_but(self, sender, eventid):
        self._apply()
        self.hide()

    def _apply(self):
        """
        Save this list of files on the node.

        Force the attribute editor to update afterwards.
        """
        attr = self.node.get_attribute("extra_uploads")
        attr.remove_all()
        for item in self.file_list_wdg.item_list:
            attr.add_string(item.get_name())
        refresh.force_ae_refresh(self.node)


def build(*args):
    """
    Pop up a window to choose extra upload files.

    Pre populate the wondow with the list of files that were stored on the
    ConductorJob item extra_uploads attribute.
    """
    node = args[0]
    win = ExtraUploadsWindow(node)
    win.show_modal()

    attr = node.get_attribute("extra_uploads")
    paths = ix.api.CoreStringArray()
    attr.get_values(paths)
    win.file_list_wdg.add_entries(paths)

    while win.is_shown():
        ix.application.check_for_events()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os


import setuptools
from PIL import Image
from setuptools.command.build_py import build_py

NAME = "cioclarisse"
DESCRIPTION = "Clarisse plugin for Conductor Cloud Rendering Platform."
URL = "https://github.com/AtomicConductor/conductor-clarisse"
EMAIL = "info@conductortech.com"
AUTHOR = "conductor"
REQUIRES_PYTHON = "~=2.7"
REQUIRED = ["ciocore>=0.2.11,<1.0.0"]
HERE = os.path.abspath(os.path.dirname(__file__))
DEV_VERSION = "dev.999"

with open(os.path.join(HERE, 'VERSION')) as version_file:
    VERSION = version_file.read().strip()

long_description = ""
with open(os.path.join(HERE, 'README.md')) as readme:
    long_description = readme.read().strip()
long_description += "\n\n## Changelog\n\n"
with open(os.path.join(HERE, 'CHANGELOG.md')) as changelog:
    long_description += changelog.read().strip()   
 


class BuildCommand(build_py):
    def run(self):
        build_py.run(self)

        if not self.dry_run:
            target_dir = os.path.join(self.build_lib, NAME)
            self.write_version_info(target_dir)

    def write_version_info(self, target_dir):
        with open(os.path.join(target_dir, "VERSION"), "w") as f:
            f.write(VERSION)

setuptools.setup(
    author=AUTHOR,
    author_email=EMAIL,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
    ],
    cmdclass={"build_py": BuildCommand},
    description=DESCRIPTION,
    install_requires=REQUIRED,
    long_description=long_description,
    long_description_content_type="text/markdown",
    name=NAME,
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    include_package_data=True,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    version=VERSION,
    zip_safe=False,
    package_data={NAME: ["scripts/*"]}
)



"""
For development mode, we can reload modules with a button push.



The button is available when the env var CONDUCTOR_MODE == "dev".
It reloads files imported by ConductorJob. Not ConductorJob itself.
"""
from cioclarisse import utils, clarisse_config
from cioclarisse import (
    attr_docs,
    debug_ui,
    dependencies,
    validation,
    environment_ui,
    extra_uploads_ui,
    frames_ui,
    instances_ui,
    job,
    validation_ui,
    notifications_ui,
    preview_ui,
    projects_ui,
    refresh,
    submission,
    submit_actions,
    task,
    upload_ui,
)
from ciocore import gpath, gpath_list, sequence

reload(utils)
reload(clarisse_config)
reload(refresh)
reload(dependencies)
reload(debug_ui)
reload(gpath_list)
reload(gpath)
reload(sequence)
reload(validation)
reload(environment_ui)
reload(extra_uploads_ui)
reload(validation_ui)
reload(frames_ui)
reload(instances_ui)
reload(job)
reload(task)
reload(notifications_ui)
reload(projects_ui)
reload(submission)
reload(submit_actions)
reload(preview_ui)
reload(upload_ui)
reload(attr_docs)



"""
Defines ConductorJob scripted class plugin. This file is the view layer.
"""
import os
import traceback

import ix
from ix.api import OfAttr

from cioclarisse.utils import ConductorError
from cioclarisse import reloader
from cioclarisse import const as k
from cioclarisse import (
    attr_docs,
    debug_ui,
    environment_ui,
    extra_uploads_ui,
    frames_ui,
    notifications_ui,
    projects_ui,
    refresh,
    submit_actions,
    upload_ui

)
from ciocore import loggeria

DEFAULT_CMD_TEMPLATE = "<temp_dir>/ciocnode <render_package> "
DEFAULT_CMD_TEMPLATE += "-image <sources> -image_frames_list <chunks> "
DEFAULT_CMD_TEMPLATE += "-tile_rendering <tiles> <tile_number> "
DEFAULT_CMD_TEMPLATE += "-license_server conductor_ilise:40500"

DEFAULT_TITLE = "My Clarisse Job"


class ConductorJob(ix.api.ModuleScriptedClassEngine):
    """
    Defines the engine that creates a ConductorJob ScriptedClass Item.
    """

    def __init__(self):
        ix.api.ModuleScriptedClassEngine.__init__(self)

    def on_action(self, action, obj, data):
        """
        Handles any button press in the Attribute Editor.

        Args:
            action (str): Name of the action is a nice version of the button label.
            obj (ConductorJob): The scripted class instance.
            data (dict): Extra information about the event
        """
        action_name = action.get_name()
        verbose_errors = obj.get_attribute("show_tracebacks").get_bool()

        # We need to wrap everything in a try/except because Clarisse
        # crashes if an exception is thrown from a scripted class. See
        # https://www.isotropix.com/user/bugtracker/363 .
        try:
            if action_name == "manage_extra_uploads":
                extra_uploads_ui.build(obj, data)
            elif action_name == "manage_extra_environment":
                environment_ui.build(obj, data)
            elif action_name == "preview":
                refresh.refresh(obj)
                submit_actions.preview(obj, data)
            elif action_name == "write_render_package":
                refresh.refresh(obj)
                submit_actions.write_render_package(obj)
            elif action_name == "submit":
                refresh.refresh(obj)
                submit_actions.submit(obj, data)
            elif action_name == "connect":
                refresh.refresh(obj, force=True)
            elif action_name == "reload":
                reload(reloader)
            else:
                pass
        except (BaseException, ValueError, AttributeError, ConductorError) as ex:
            if verbose_errors:
                ix.log_error(traceback.format_exc())
            else:
                ix.log_error(ex.message)

    def on_attribute_change(self, obj, attr, *_):
        """
        Handles attributes changing value.

        Args:
            obj (ConductorJob): The scripted class instance.
            attr (OfAttr): The attribute that changed.
        """

        attr_name = attr.get_name()
        verbose_errors = obj.get_attribute("show_tracebacks").get_bool()
        try:
            if attr_name == "conductor_project_name":
                projects_ui.handle_project(obj, attr)
            elif attr_name == "use_custom_frames":
                frames_ui.handle_use_custom_frames(obj, attr)
            elif attr_name == "use_scout_frames":
                frames_ui.handle_use_scout_frames(obj, attr)
            elif attr_name == "custom_frames":
                frames_ui.handle_custom_frames(obj, attr)
            elif attr_name == "scout_frames":
                frames_ui.handle_scout_frames(obj, attr)
            elif attr_name == "chunk_size":
                frames_ui.handle_chunk_size(obj, attr)
            elif attr_name == "images_and_layers":
                frames_ui.handle_images(obj, attr)
            elif attr_name == "notify":
                notifications_ui.notify_changed(obj, attr)
            elif attr_name == "email_addresses":
                notifications_ui.handle_email_addresses(obj, attr)
            elif attr_name == "conductor_log_level":
                debug_ui.handle_log_level(obj, attr)
            elif attr_name == "local_upload":
                upload_ui.handle_local_upload(obj, attr)
            else:
                pass
        except (BaseException, ValueError, AttributeError, ConductorError) as ex:
            if verbose_errors:
                ix.log_error(traceback.format_exc())
            else:
                ix.log_error(ex.message)

    def declare_attributes(self, s_class):
        """
        Declares all attributes and actions.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        self.declare_actions(s_class)
        self.declare_general_attributes(s_class)
        self.declare_frames_attributes(s_class)
        self.declare_machines_attributes(s_class)
        self.declare_upload_attributes(s_class)
        self.declare_environment_attributes(s_class)
        self.declare_task_attributes(s_class)
        self.declare_packaging_attributes(s_class)
        self.declare_notification_attributes(s_class)
        self.declare_dev_attributes(s_class)

        attr_docs.set_docs(s_class)

    def declare_dev_attributes(self, s_class):
        """
        Declares developer attributes.

        Sets up log level and traceback behaviour. Also provides a reload button
        for Conductor devs by setting an env var.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        dev_visible = os.environ.get("CONDUCTOR_MODE") == "dev"

        if dev_visible:
            self.add_action(s_class, "reload", "development")
        self.add_action(s_class, "write_render_package", "debug")

        attr = s_class.add_attribute(
            "conductor_log_level",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "debug",
        )
        attr.set_long(1)
        for i, level in enumerate(loggeria.LEVELS):
            attr.add_preset(level, str(i))

        attr = s_class.add_attribute(
            "show_tracebacks",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "debug",
        )
        attr.set_bool(False)

    def declare_packaging_attributes(self, s_class):
        """
        Declares render package attributes.

        Render package is simply a project file that may have been manipulated
        in preparation for the render.

        localize_contexts: Like Maya's import references. Useful if renders fail
        due to the complexity of dispatching nested references.


        clean_up_render_package: Remove the render package after submission.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """

        attr = s_class.add_attribute(
            "clean_up_render_package",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "packaging",
        )
        attr.set_bool(True)

    def declare_actions(self, s_class):
        """
        Creates connect, preview, and submit buttons.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        self.add_action(s_class, "connect", "actions")
        # self.add_action(s_class, "validate", "actions")
        self.add_action(s_class, "preview", "actions")
        self.add_action(s_class, "submit", "actions")

    def declare_general_attributes(self, s_class):
        """
        Creates commonly accessed attributes.

        Job title, project, and images to be rendered.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        attr = s_class.add_attribute(
            "title",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "general",
        )
        attr.set_string(DEFAULT_TITLE)

        attr = s_class.add_attribute(
            "images_and_layers",
            OfAttr.TYPE_REFERENCE,
            OfAttr.CONTAINER_LIST,
            OfAttr.VISUAL_HINT_DEFAULT,
            "general",
        )
        filters = ix.api.CoreStringVector(0)
        filters.add("Layer")
        filters.add("Image")
        attr.set_object_filters(filters)

        attr = s_class.add_attribute(
            "conductor_project_name",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "general",
        )
        attr.set_long(0)
        attr.add_preset(k.NOT_CONNECTED,  "0")

        attr = s_class.add_attribute(
            "last_project",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "general",
        )
        attr.set_hidden(True)

        attr = s_class.add_attribute(
            "clarisse_version",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "general",
        )
        attr.set_long(0)

    def declare_frames_attributes(self, s_class):
        """
        Creates attributes related to the frame range.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """

        attr = s_class.add_attribute(
            "use_custom_frames",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_bool(False)

        attr = s_class.add_attribute(
            "custom_frames",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_hidden(True)
        attr.set_string("2,4-8,10-50x2")

        attr = s_class.add_attribute(
            "chunk_size",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_long(5)
        attr.set_numeric_range_min(1)

        attr = s_class.add_attribute(
            "use_scout_frames",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_bool(False)

        attr = s_class.add_attribute(
            "scout_frames",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_hidden(True)
        attr.set_string("2,4-8,10-50x2")

        attr = s_class.add_attribute(
            "tiles",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_long(1)
        for i in range(1, 11):
            [attr.add_preset("{}x{}={}".format(i, i, i * i), str(i))]

        attr = s_class.add_attribute(
            "frames_info",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "frames",
        )
        attr.set_read_only(True)
        attr.set_string("- Please click connect -")

    def declare_machines_attributes(self, s_class):
        """
        Creates attributes to define the render node spec.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        attr = s_class.add_attribute(
            "preemptible",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "machines",
        )
        attr.set_bool(True)

        attr = s_class.add_attribute(
            "instance_type",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "machines",
        )
        attr.set_long(0)
        attr.add_preset(k.NOT_CONNECTED,  "0")

        attr = s_class.add_attribute(
            "retries",
            OfAttr.TYPE_LONG,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "machines",
        )
        attr.set_long(3)

        attr = s_class.add_attribute(
            "last_instance_type",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "machines",
        )
        attr.set_hidden(True)

    def declare_upload_attributes(self, s_class):
        """
        Creates attributes to specify options for dependency scanning.

        No-scan: use only cached uploads.

        Glob-scan: If filenames contain "##" or <UDIM>. They will be globbed.

        Smart-scan If filenames contain "##" or $<?>F the hashes will be
        replaced with frame numbers that are being used.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """

        attr = s_class.add_attribute(
            "local_upload",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "upload",
        )
        attr.set_bool(True)

        attr = s_class.add_attribute(
            "upload_only",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "upload",
        )
        attr.set_bool(False)

        self.add_action(s_class, "manage_extra_uploads", "upload")

        attr = s_class.add_attribute(
            "extra_uploads",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_LIST,
            OfAttr.VISUAL_HINT_DEFAULT,
            "cached_upload_list",
        )
        attr.set_read_only(True)

    def declare_task_attributes(self, s_class):
        """
        Creates task command template.

        By default it contain tokens that will be replaced while
        generating tasks.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        attr = s_class.add_attribute(
            "task_template",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "task",
        )
        attr.set_string(DEFAULT_CMD_TEMPLATE)

    def declare_environment_attributes(self, s_class):
        """
        Sets up any extra environment.

        Vars set here will be merged with the package environments.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        self.add_action(s_class, "manage_extra_environment", "environment")

        attr = s_class.add_attribute(
            "extra_environment",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_LIST,
            OfAttr.VISUAL_HINT_DEFAULT,
            "environment",
        )
        attr.set_read_only(True)

    def declare_notification_attributes(self, s_class):
        """
        Sets email addresses to be notified on job completion.

        Args:
            s_class (ScriptedClass):  The scripted class.
        """
        attr = s_class.add_attribute(
            "notify",
            OfAttr.TYPE_BOOL,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "notifications",
        )
        attr.set_bool(False)

        attr = s_class.add_attribute(
            "email_addresses",
            OfAttr.TYPE_STRING,
            OfAttr.CONTAINER_SINGLE,
            OfAttr.VISUAL_HINT_DEFAULT,
            "notifications",
        )
        attr.set_read_only(True)



import os
import sys
from shutil import copy2
 
CIO_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
CONFIG_FILENAME = "clarisse.env"
PLATFORM = sys.platform

def main():
    sys.stdout.write("Running Clarisse post install script\n")

    if not PLATFORM in ["darwin", "win32", "linux"]:
        sys.stderr.write("Unsupported platform: {}".format(PLATFORM))
        sys.exit(1)
    transform_env_files()
    sys.stdout.write("Completed Clarisse post install script\n")

    sys.exit(0)


def transform_env_files():
    """Find env files below the Clarisse dir and add a Conductor location variable."""
    if PLATFORM == "darwin":
        root = os.path.expanduser(os.path.join("~","Library","Preferences","Isotropix","Clarisse"))
    elif PLATFORM == "linux":
        root = os.path.expanduser(os.path.join("~",".isotropix","clarisse"))
    else:  # windows
        root =  os.path.expanduser(os.path.join("~","AppData", "Roaming", "Isotropix", "Clarisse"))
    sys.stdout.write("Found Clarisse preferences folder: {}\n".format(root))
    transform_tree(root)


def transform_tree(root):
    """Walk the filesystem and transform env files."""
    root, dirs, filenames = next(os.walk(root), (None,[], []))
    for fn in filenames:
        if fn == CONFIG_FILENAME:
            transform_env_file(os.path.join(root, fn))
    if dirs:
        for dr in dirs:
            transform_tree(os.path.join(root, dr))

def transform_env_file(env_file):
    sys.stdout.write("Write CIO additions to file: {}\n".format(env_file))
    lines = []
    real_env_file = os.path.realpath(env_file)
    bak = "{}.bak".format(real_env_file)

    with open(real_env_file) as fn:
        for line in fn:
            if not line.startswith("CIO_DIR"):
                lines.append(line.rstrip())

    sys.stdout.write("Backup env file to: {}\n".format(real_env_file))
    os.rename(real_env_file, bak)
    
    lines.append("CIO_DIR={}".format(CIO_DIR))
    try:
        with open(real_env_file, "w") as fn:
            for line in lines:
                fn.write("{}\n".format(line))
    except BaseException:
        sys.stderr.write("Couldn't write CIO_DIR path to : {}\n".format(env_file))
        sys.stderr.write("Please edit the file add it yourself: CIO_DIR={}\n".format(CIO_DIR))
        os.rename(bak, real_env_file)
        return
        
    sys.stdout.write("Wrote Conductor additions to : {}\n".format(env_file))

if __name__ == '__main__':
    main()
